/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.dom.tester;

import com.ibm.dom.util.Digest;
import com.ibm.xml.parser.TXDocument;
import com.ibm.xml.parser.TXElement;
import com.ibm.xml.sax.DigestFilter;
import com.ibm.xml.sax.NamespaceAdapter;
import com.sun.xml.tree.XmlDocument;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.security.MessageDigest;
import java.util.Enumeration;
import java.util.Vector;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.AttributeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.HandlerBase;
import org.xml.sax.InputSource;
import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class TestDomhash {

    // ================================================================

    static class Item {
        String src;
        String digest;
        String domhash;
        public Item(String s, String dig, String dom) {
            this.src = s.intern();
            this.digest = dig.intern();
            this.domhash = dom.intern();
        }
    }

    // ================================================================

    static class DigestTable extends HandlerBase {
        String algorithm;
        Vector items = new Vector();
        int maxsrc = 1;
        public void startElement(String name, AttributeList attributes) throws SAXException {
            if (name.equals("digests")) {
                this.algorithm = attributes.getValue("algorithm");
            } else if (name.equals("item")) {
                String src = attributes.getValue("src");
                if (src.length() > this.maxsrc)  this.maxsrc = src.length();
                this.items.addElement(new Item(src,
                                               attributes.getValue("digest"),
                                               attributes.getValue("domhash")));
            }
        }

        void fromDOM(Document doc) {
            Element root = doc.getDocumentElement();
            if (!root.getNodeName().equals("digests"))
                throw new RuntimeException("The root element is not \"digests\".");
            this.algorithm = root.getAttribute("algorithm");
            for (Node ch = root.getFirstChild();  ch != null;
                 ch = ch.getNextSibling()) {
                if (ch.getNodeType() == Node.ELEMENT_NODE) {
                    if (!ch.getNodeName().equals("item"))
                        throw new RuntimeException("The element is not \"item\".");
                    String domhash = ((Element)ch).getAttribute("domhash");
                    if (domhash == null)  domhash = ""; // For a bug of older XML4J
                    String src = ((Element)ch).getAttribute("src");
                    if (src.length() > this.maxsrc)  this.maxsrc = src.length();
                    this.items.addElement(new Item(src,
                                                   ((Element)ch).getAttribute("digest"),
                                                   domhash.length() == 0 ? null : domhash));
                }
            }
        }
    }

    // ================================================================

    static class Stderr implements ErrorHandler {
        public void warning(SAXParseException exception) throws SAXException {
            //System.out.println("WARNING: "+exception.getMessage());
        }
        public void error(SAXParseException exception) throws SAXException {
            System.out.println("ERROR: "+exception.getMessage());
        }
        public void fatalError(SAXParseException exception) throws SAXException {
            //System.out.println("FATAL ERROR: "+exception.getMessage());
            throw exception;
        }
    }

    // ================================================================

    static class EntityResolverImpl implements EntityResolver {
        String url;
        byte[] cont;
        EntityResolverImpl(String url, byte[] cont) {
            this.url = url;
            this.cont = cont;
        }
        public InputSource resolveEntity(String publicId, String systemId) {
            if (systemId.equals(url)) {
                InputSource is = new InputSource(new ByteArrayInputStream(this.cont));
                is.setPublicId(publicId);
                is.setSystemId(systemId);
                return is;
            }
            return null;
        }
    }

    // ================================================================

    static String toString(byte[] ab) {
        final String tab = "0123456789ABCDEF";
        StringBuffer sb = new StringBuffer(ab.length*2);
        for (int i = 0;  i < ab.length;  i ++) {
            sb.append(tab.charAt((ab[i]>>4)&0x0f));
            sb.append(tab.charAt(ab[i]&0x0f));
        }
        return sb.toString().intern();
    }

    static final int T_SAX = 0;
    static final int T_XML4J1 = 1;
    static final int T_XML4J2 = 2;
    static final int T_PROJX = 3;
    static final int T_OPENX = 4;

    static Document parseDocument(int type, String url) throws IOException, SAXException {
        Document doc = null;
        switch (type) {
          case T_XML4J1:
            doc = XML4J1.parseDocument(url);
            break;
          case T_XML4J2:
            doc = XML4J2.parseDocument(url);
            break;
          case T_PROJX:
            doc = ProjectX.parseDocument(url);
            break;
          case T_OPENX:
            doc = OpenXML.parseDocument(url);
            break;
          default:
        }
        if (doc == null || doc.getDocumentElement() == null)
            throw new RuntimeException("Null document.");
        return doc;
    }
    static class XML4J1 {
        static Document parseDocument(String url) throws IOException {
            com.ibm.xml.parser.Parser xml4jp = new com.ibm.xml.parser.Parser(url);
            xml4jp.setProcessNamespace(true);
            xml4jp.setWarningNoXMLDecl(false);
            return xml4jp.readStream(xml4jp.getInputStream(url, null, url));
        }
    }
    static class XML4J2 {
        static Document parseDocument(String url) throws IOException, SAXException {
            com.ibm.xml.parsers.DOMParser domp = new com.ibm.xml.parsers.DOMParser();
            domp.parse(url);
            return domp.getDocument();
        }
    }
    static class ProjectX {
        static Document parseDocument(String url) throws SAXException, IOException {
            return XmlDocument.createXmlDocument(url);
        }
    }
    static class OpenXML {
        static Document parseDocument(String url) throws IOException {
            org.openxml.source.Source oxsource = org.openxml.DOMFactory.newSource();
            oxsource.setURI(url);
            return oxsource.getDocument();
        }
    }

    // ================================================================

    interface RawDigestMatchingListener {
        void handleUnmatchedDigest(String expected, String result);
    }

    static String check(Item item, Config conf, MessageDigest md,
                        String digestAlgorithm,
                        Parser parser, ErrorHandler stderr,
                        RawDigestMatchingListener listener) {
        String domhash = null;
        try {
            String targetDocumentString = conf.base + item.src;
            URL targetDocument = new URL(targetDocumentString);
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            InputStream is = targetDocument.openStream();
            byte[] buf = new byte[1024];
            int len;
            while ((len = is.read(buf)) >= 0) {
                baos.write(buf, 0, len);
            }
            is.close();
            baos.close();
            byte[] dococtet = baos.toByteArray();

            md.reset();
            String realdigest = toString(md.digest(dococtet));
            if (!item.digest.equals(realdigest)) {
                System.out.println(item.src+" has been modified: table="
                                   +item.digest+" realdata="+realdigest);
                if (listener != null)
                    listener.handleUnmatchedDigest(item.digest, realdigest);
            }

            switch (conf.type) {
              case T_SAX:
                NamespaceAdapter nsa = new NamespaceAdapter(parser);
                DigestFilter digest = new DigestFilter(md);
                nsa.setDocumentHandler(digest);
                nsa.setErrorHandler(stderr);
                nsa.setEntityResolver(new EntityResolverImpl(targetDocumentString, dococtet));
                nsa.parse(targetDocumentString);
                domhash = toString(digest.getDigest());
                break;

              default:
                Document doc = parseDocument(conf.type, targetDocumentString);
                if (conf.type == T_XML4J1) {
                    ((TXDocument)doc).setDigestAlgorithm(digestAlgorithm);
                    domhash = toString(((TXElement)doc.getDocumentElement()).getDigest());
                } else {
                    domhash = toString(Digest.getDigest(md, doc));
                }
                break;
            }
        } catch (Exception ex) {
            System.err.println(item.src+": FAILED: "+ex);
        }
        return domhash;
    }

    static class Config {
        int type = T_SAX;
        String parserClass = null;
        String table = null;
        String base = null;

        Config() {
        }

        Config(String[] argv) {
            for (int i = 0;  i < argv.length;  i ++) {
                if (argv[i].charAt(0) == '-') {
                    if (argv[i].equals("-xml4j1")) {
                        this.type = T_XML4J1;
                    } else if (argv[i].equals("-xml4j2")) {
                        this.type = T_XML4J2;
                    } else if (argv[i].equals("-projx")) {
                        this.type = T_PROJX;
                    } else if (argv[i].equals("-openx")) {
                        this.type = T_OPENX;
                    } else {
                        System.err.println("Unknown option: "+argv[i]);
                    }
                } else {
                    if (this.type == T_SAX && this.parserClass == null) {
                        this.parserClass = argv[i];
                    } else if (this.table == null) {
                        this.table = argv[i];
                    } else if (this.base == null) {
                        this.base = argv[i];
                        if (this.base.charAt(this.base.length()-1) != '/')
                            this.base = this.base + "/";
                    } else {
                        System.err.println("???: "+argv[i]);
                    }
                }
            }
        }
    }

    static void printUsage() {
        System.err.println("<SAX Parser class>:");
        System.err.println("\tcom.ibm.xml.parser.SAXDriver");
        System.err.println("\tcom.ibm.xml.parsers.SAXParser");
        System.err.println("\tcom.ibm.xml.parsers.ValidatingSAXParser");
        System.err.println("\tcom.jclark.xml.sax.Driver");
        System.err.println("\tcom.microstar.xml.SAXDriver");
        System.err.println("\tcom.sun.xml.parser.Parser");
        System.err.println("\tcom.sun.xml.parser.ValidatingParser");
        System.err.println("\tetc.");
        System.err.println("Options:");
        System.err.println("\t-xml4j1\tUse built-in getDigest() method");
        System.err.println("\t-xml4j2\tUse com.ibm.dom.util.Digest against XML4J2 DOM");
        System.err.println("\t-projx\tUse com.ibm.dom.util.Digest against Project X DOM");
        System.err.println("\t-openx\tUse com.ibm.dom.util.Digest against OpenXML DOM");
    }

    public static void main(String[] argv) {
        Config conf = new Config(argv);
        if (conf.base == null) {
            System.err.println("java com.ibm.dom.util.TestDomhash <SAX Parser class> <table URL> <documents base URL>");
            System.err.println("java com.ibm.dom.util.TestDomhash [options] <table URL> <documents base URL>");
            printUsage();
            System.exit(1);
        }

        try {
            DigestTable tab = new DigestTable();
            Parser parser = null;
            ErrorHandler stderr = null;
            if (conf.type == T_SAX) {
                parser = (Parser)Class.forName(conf.parserClass).newInstance();
                parser.setDocumentHandler(tab);
                parser.setErrorHandler(stderr = new Stderr());
                parser.parse(conf.table);
            } else {
                tab.fromDOM(parseDocument(conf.type, conf.table));
            }

            MessageDigest md = MessageDigest.getInstance(tab.algorithm);
            Enumeration en = tab.items.elements();
            while (en.hasMoreElements()) {
                Item item = (Item)en.nextElement();
                if (item.domhash != null) {
                    String domhash = check(item, conf, md, tab.algorithm,
                                           parser, stderr, null);
                    if (domhash == null) {
                    } else if (!item.domhash.equals(domhash)) {
                        System.out.println(item.src+": FAILED: Mismatched DOMHASH: table="
                                           +item.domhash+" realdata="+domhash);
                    } else {
                        System.out.println(item.src+": OK.");
                    }
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
