/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.dom.util;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.security.MessageDigest;
import java.util.Hashtable;
import java.util.Vector;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 * @see com.ibm.xml.parser.MakeDigestVisitor
 * @see com.ibm.xml.sax.DigestFilter
 */
public class Digest {
    public static final String S_XMLNS = "xmlns";
    public static final String S_XMLNAMESPACEURI = "http://www.w3.org/XML/1998/namespace";

    private Digest() {
    }

    public static byte[] getDigest(MessageDigest md, Node node) {
        if (node == null)
            throw new IllegalArgumentException("The parameter is null.");
        md.reset();
        byte[] result = null;
        int type = node.getNodeType();
        switch (type) {
          case Node.DOCUMENT_NODE:
            return Digest.getDigest(md, ((Document)node).getDocumentElement());

          case Node.ELEMENT_NODE:
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            Digest.writeInt(baos, Node.ELEMENT_NODE); // Node type
            String str = Digest.createExpandedName(node, node);
            for (int i = 0;  i < str.length();  i ++) {
                int ch = str.charAt(i);
                baos.write(ch >> 8);
                baos.write(ch & 0xff);
            }
            baos.write(0);
            baos.write(0);

            NamedNodeMap nnm = node.getAttributes();
            int atlen = nnm.getLength();
            if (atlen == 0) {                   // Number of attributes
                Digest.writeInt(baos, 0);
            } else {
                int lenWithoutXmlns = 0;
                Hashtable hash = new Hashtable(atlen);
                String[] as = new String[atlen]; // Attributes orderred by expanded name
                for (int i = 0;  i < atlen;  i ++) {
                    Node attr = nnm.item(i);
                    String qname = attr.getNodeName();
                    if (!qname.equals("xmlns") && !qname.startsWith("xmlns:")) {
                        String expandedName = Digest.createExpandedName(attr, node);
                        hash.put(expandedName, qname);
                        as[lenWithoutXmlns++] = expandedName;
                    }
                }
                Digest.writeInt(baos, lenWithoutXmlns);
                Digest.heapSort(as, lenWithoutXmlns);
                try {
                    for (int i = 0;  i < lenWithoutXmlns;  i ++) {
                        Attr a = ((Element)node).getAttributeNode((String)hash.get(as[i]));
                        baos.write(Digest.getDigest(md, a, ((Element)node)));
                    }
                } catch (IOException ioe) {
                                                // Never thrown
                }
            }
                    
                                                // Collect children digests
            Vector v = new Vector();
            StringBuffer sb = new StringBuffer();
            for (Node n = Digest.getFirstWithoutReference(node);
                 n != null;
                 n = Digest.getNextWithoutReference(n)) {
                byte[] digest0 = null;

                // This condition isn't n.getNodeType() == Node.TEXT_NODE.
                // Includes CDATA
                if (n instanceof Text) {
                    sb.append(n.getNodeValue());
                    Node m;
                    while ((m = Digest.getNextWithoutReference(n)) != null
                           && m instanceof Text) {
                        n = m;
                        sb.append(n.getNodeValue());
                    }
                    if (0 < sb.length()) {
                        digest0 = Digest.getTextDigest(md, sb.toString());
                        sb.setLength(0);
                    }
                } else if (n.getNodeType() != Node.COMMENT_NODE) {
                    digest0 = Digest.getDigest(md, n);
                }
                if (digest0 != null)
                    v.addElement(digest0);
            }
            Digest.writeInt(baos, v.size());    // Number of children
            try {
                for (int i = 0;  i < v.size();  i ++)
                    baos.write((byte[])v.elementAt(i)); // Children digests
                baos.close();
            } catch (IOException io) {
                                                // Never thrown
            }
            

            md.reset();
            md.update(baos.toByteArray());
            result = md.digest();
            break;

          case Node.PROCESSING_INSTRUCTION_NODE:
            Digest.digestInt(md, type);
            Digest.digestString(md, node.getNodeName());
            md.update((byte)0);
            md.update((byte)0);
            Digest.digestString(md, node.getNodeValue());
            result = md.digest();
            break;

          case Node.TEXT_NODE:
          case Node.CDATA_SECTION_NODE:
            Digest.digestInt(md, Node.TEXT_NODE);
            Digest.digestString(md, node.getNodeValue());
            result = md.digest();
            break;
        }
        return result;
    }

    public static byte[] getDigest(MessageDigest md, Attr attr, Element owner) {
        String qname = attr.getNodeName();
        if (qname.equals("xmlns") || qname.startsWith("xmlns:"))
            return null;
        md.reset();
        Digest.digestInt(md, Node.ATTRIBUTE_NODE);
        Digest.digestString(md, Digest.createExpandedName(attr, owner));
        md.update((byte)0);
        md.update((byte)0);
        Digest.digestString(md, attr.getValue());
        return md.digest();
    }

    private static byte[] getTextDigest(MessageDigest md, String str) {
        md.reset();
        Digest.digestInt(md, Node.TEXT_NODE);
        Digest.digestString(md, str);
        return md.digest();
    }

    // ================================================================
    private static void writeInt(ByteArrayOutputStream baos, int i) {
        baos.write((i >> 24) & 0xff);
        baos.write((i >> 16) & 0xff);
        baos.write((i >> 8) & 0xff);
        baos.write(i & 0xff);
    }

    // ================================================================

    /**
     *
     * @exceptin LibraryException node is neither <CODE>Element</CODE> nor <CODE>Attr</CODE>.
     */
    public static String createExpandedName(Node node, Node element) {
        String ret;
        if (node.getNodeType() == Node.ELEMENT_NODE) {
            String name = node.getNodeName();
            int index = name.indexOf(':');
            String prefix = index <= 0 ? "" : name.substring(0, index);
            String uri = Digest.getNamespaceForPrefix(prefix, element);
            if (uri == null || uri.length() == 0) {
                ret = name;
            } else {
                ret = uri+":"+name.substring(index+1);
            }
        } else if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
            String name = node.getNodeName();
            int index = name.indexOf(':');
            if (index <= 0) {
                ret = name;
            } else {
                ret = Digest.getNamespaceForPrefix(name.substring(0, index), element)
                    +":"+name.substring(index+1);
            }
        } else {
            throw new IllegalArgumentException("Requires Element or Attr node.");
        }
        return ret;
    }

    /**
     * Returns a namespace URI for specified prefix.
     * <TABLE border="1">
     *   <TR><TD>namespace URI String<TD>for valid prefix or no prefix with default namespace</TR>
     *   <TR><TD>""<TD>for no prefix and no default namespace</TR>
     *   <TR><TD>null<TD>for undefined prefix</TR>
     * </TABLE>
     * @param prefix Namespace prefix; null isn't allowed.
     */
    public static String getNamespaceForPrefix(String prefix, Node current) {
        if (prefix.equals("xml"))
            return S_XMLNAMESPACEURI;
        if (prefix.equals(S_XMLNS))
            return prefix;                      // ???
        String attributeName = prefix.length() == 0 ? S_XMLNS : ("xmlns:"+prefix);
        Node element = current;
        do {
                     // This verbose condition is for a bug of Sun's Project X ea2.
            Node attr = element.getAttributes().getLength() == 0 ? null : ((Element)element).getAttributeNode(attributeName);
            if (attr != null)
                return attr.getNodeValue();
            while (true) {
                element = element.getParentNode();
                if (element == null)  return attributeName == S_XMLNS ? "" : null;
                int type = element.getNodeType();
                if (type == Node.ENTITY_REFERENCE_NODE) {
                } else if (type == Node.ELEMENT_NODE) {
                    break;
                } else
                    return attributeName == S_XMLNS ? "" : null;
            }
        } while (element != null);              // This condition has no sense.
        return null;                            // Not reached
    }

    // ================================================================
    /*
     * Returns the Node preceding specified Node. This method ignore EntityReference nodes.
     * If there is no such Node, <var>null</var> is returned.
     *
     * @return          The child Node preceding specified node,
     *                    or <var>null</var> if the node is the last sibling.
     * @see #getNextWithoutReference
     */
    public static Node getPreviousWithoutReference(Node node) {
        Node ret = node.getPreviousSibling();
        if (ret == null) {
            Node parent = node.getParentNode();
            if (parent == null)   return null;
            if (parent.getNodeType() == Node.ENTITY_REFERENCE_NODE)
                ret = getPreviousWithoutReference(parent);
        } else {                                // not null
            while (ret.getNodeType() == Node.ENTITY_REFERENCE_NODE) {
                Node last = ret.getLastChild();
                if (last == null) {
                    ret = getPreviousWithoutReference(ret);
                    break;
                } else
                    ret = last;
            }
        }
        return ret;
    }

    /*
     * Returns the Node following specified Node. This method ignore EntityReference nodes.
     * If there is no such Node, <var>null</var> is returned.
     *
     * @return          The child Node following specified node,
     *                    or <var>null</var> if the node is the last sibling.
     * @see #getPreviousWithoutReference
     */
    public static Node getNextWithoutReference(Node node) {
        Node ret = node.getNextSibling();
        if (ret == null) {
            Node parent = node.getParentNode();
            if (parent == null)   return null;
            if (parent.getNodeType() == Node.ENTITY_REFERENCE_NODE)
                ret = getNextWithoutReference(parent);
        } else {                                // not null
            while (ret.getNodeType() == Node.ENTITY_REFERENCE_NODE) {
                Node first = ret.getFirstChild();
                if (first == null) {
                    ret = getNextWithoutReference(ret);
                    break;
                } else
                    ret = first;
            }
        }
        return ret;
    }

    /**
     *
     * @see #getLastWithoutReference
     */
    public static Node getFirstWithoutReference(Node node) {
        Node ret = node.getFirstChild();
        while (ret != null && ret.getNodeType() == Node.ENTITY_REFERENCE_NODE) {
            Node first = ret.getFirstChild();
            if (first == null) {
                ret = getNextWithoutReference(ret);
                break;
            }
            ret = first;
        }
        return ret;
    }

    /**
     * @see #getFirstWithoutReference
     */
    public static Node getLastWithoutReference(Node node) {
        Node ret = node.getLastChild();
        while (ret != null && ret.getNodeType() == Node.ENTITY_REFERENCE_NODE) {
            Node last = ret.getLastChild();
            if (last == null) {
                ret = getPreviousWithoutReference(ret);
                break;
            }
            ret = last;
        }
        return ret;
    }

    // ================================================================
    /**
     * Writes 32bit integer in network byte order.
     */
    private static void digestInt(MessageDigest md, int i) {
        md.update((byte)((i >> 24) & 0xff));
        md.update((byte)((i >> 16) & 0xff));
        md.update((byte)((i >> 8) & 0xff));
        md.update((byte)(i & 0xff));
    }

    /**
     * Writes String in big endian UTF-16.
     */
    private static void digestString(MessageDigest md, String str) {
        for (int i = 0;  i < str.length();  i ++) {
            int ch = str.charAt(i);
            md.update((byte)(ch >> 8));
            md.update((byte)(ch & 0xff));
        }
    }

    // ================================================================
    /**
     * Sort String array.
     */
    private static void heapSort(String[] pd, int length) {
        int i;
        for (i = length/2;  i >= 0;  i--) {     // Make heap
            fall(pd, length, i);
        }
        for (i = length-1;  i > 0;  i--) {
            String t = pd[0];
            pd[0] = pd[i];
            pd[i] = t;
            fall(pd, i, 0);
        }
    }
    private static void fall(String[] pd, int n, int i) {
        int j = 2*i+1;
        if (j < n) {                            // left exists
            if (j+1 < n) {                      // right exists too
                                                // j: bigger
                if (0 > pd[j].compareTo(pd[j+1]))
                    j = 2*i+2;
            } else {                            // only left
            }
            if (0 > pd[i].compareTo(pd[j])) {
                                                // the child is bigger
                String t = pd[i];
                pd[i] = pd[j];
                pd[j] = t;
                fall(pd, n, j);
            }
        }
    }
}
