/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.dom.util;

import java.io.IOException;
import java.io.Writer;
import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;

/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author TAMURA Kent &lt;tkent@jp.ibm.com&gt;
 * @see com.ibm.dom.util.TreeTraversal
 * @see com.ibm.xml.parser.ToXMLStringVisitor
 */
public class ToXMLVisitor extends NOOPVisitor {
    protected Writer writer = null;
    protected String encoding = null;
    protected boolean isPrintNonSpecifiedAttributes = true;

    /**
     * Constructor for customized encoding.
     * @param writer        The character output stream to use. This MUST be UTF-8 or UTF-16 Writer.
     * @param encoding      Java character encoding in use by <VAR>writer</VAR>.
     */
    public ToXMLVisitor(Writer writer, String encoding) {
        this.writer = writer;
        this.encoding = encoding;
    }

    /**
     * Constructor for default encoding.
     * @param writer  The character output stream to use. This MUST be UTF-8 or UTF-16 Writer.
     */
    public ToXMLVisitor(Writer writer) {
        this(writer, null);
    }

    /**
     * Sets whether this prints non-specified default attributes appended by DTD.
     * <P>By default, non-specified attributes are printed.
     * @see org.w3c.dom.Attr#getSpecified
     */
    public void setPrintNonSpecifiedAttributes(boolean print) {
        this.isPrintNonSpecifiedAttributes = print;
    }

    /**
     * Returns whether this prints non-specified default attributes appended by DTD.
     */
    public boolean getPrintNonSpecifiedAttributes() {
        return this.isPrintNonSpecifiedAttributes;
    }

    /**
     * Creates a string representation of the specified <var>document</var> Node in XML format.
     * @param   document    Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitDocumentPre(Document document) throws Exception {
        this.writer.write("<?xml version=\"1.0\"?>");
    }

    /**
     * Flush the writer.
     * @param   document    Not used.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitDocumentPost(Document document) throws Exception {
        this.writer.flush();
    }

    /**
     * Creates a string representation of the start of the specified <var>element</var> Node
     * and its associated attributes in XML format.
     * @param   element     Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitElementPre(Element element) throws Exception {
        this.writer.write("<");
        this.writer.write(element.getTagName());
        NamedNodeMap nnm = element.getAttributes();
        synchronized (nnm) {
            for (int i = 0;  i < nnm.getLength();  i ++) {
                Attr attr = (Attr)nnm.item(i);
                if (this.isPrintNonSpecifiedAttributes || attr.getSpecified())
                    visitAttributePre(attr);
            }
        }
        if (element.hasChildNodes()) {
            this.writer.write(">");
        } else {
            this.writer.write("/>");
        }
    }

    /**
     * Creates a string representation of the end of the specified <var>element</var> Node
     * in XML format.
     * @param   element     Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitElementPost(Element element) throws Exception {
        if (element.hasChildNodes()) {
            this.writer.write("</");
            this.writer.write(element.getTagName());
            this.writer.write(">");
        }
    }

    /**
     * Creates a string representation of the specified <var>attribute</var> Node in XML
     * format.
     * @param   attribute   Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitAttributePre(Attr attribute) throws Exception {
        this.writer.write(" ");
        this.writer.write(attribute.getName());
        this.writer.write("=\"");
        for (Node child = attribute.getFirstChild();  child != null;
             child = child.getNextSibling()) {
            int type = child.getNodeType();
            if (type == Node.TEXT_NODE) {
                this.escapedWrite(this.writer, child.getNodeValue(), "<&\"", this.encoding);
            } else if (type == Node.ENTITY_REFERENCE_NODE) {
                this.writer.write("&");
                this.writer.write(child.getNodeName());
                this.writer.write(";");
            } else {
                //???
            }
        }
        this.writer.write("\"");
    }

    /**
     * Creates a string representation of the specified <var>pi</var> Node in XML format.
     * @param   pi          Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitProcessingInstructionPre(ProcessingInstruction pi) throws Exception {
        this.writer.write("<?");
        this.writer.write(pi.getNodeName());
        String data = pi.getData();
        if (pi.getData().length() > 0) {
           this.writer.write(" ");
           this.writer.write(pi.getData());
        }
        this.writer.write("?>");
    }

    /**
     * Creates a string representation of the specified <var>comment</var> Node in XML format.
     * @param   comment     Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitCommentPre(Comment comment) throws Exception {
        this.writer.write("<!--");
        this.writer.write(comment.getData());
        this.writer.write("-->");
    }

    /**
     * Creates a string representation of the specified <var>text</var> Node in XML format.
     * @param   text        Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitTextPre(Text text) throws Exception {
        this.escapedWrite(this.writer, text.getData(), "<&", this.encoding);
    }

    /**
     * Creates a string representation of the specified <var>cdata</var> Node in XML format.
     * @param   cdata        Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitCDATASectionPre(CDATASection cdata) throws Exception {
        this.writer.write("<![CDATA[");
        this.writer.write(cdata.getData());
        this.writer.write("]]>");
    }

    /**
     * Creates a string representation of the specified <var>doctype</var> Node
     * in XML format.
     * @param   doctype         Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitDocumentTypePre(DocumentType doctype) throws Exception {
        this.writer.write("<!DOCTYPE ");
        this.writer.write(doctype.getName());
        this.writer.write(">");
    }

    /**
     * Creates a string representation of the specified <var>ref</var> Node in XML format.
     * @param   ref    Node toString in XML format.
     * @exception Exception   Thrown if this Node can not be visitted because of an invalid character output stream.
     */
    public void visitEntityReference(EntityReference ref) throws Exception {
        this.writer.write("&");
        this.writer.write(ref.getNodeName());
        this.writer.write(";");
        throw new TreeTraversal.ToNextSiblingTraversalException();
    }

    // ================================================================

    /**
     * Writes the specified <var>string</var> to specified Writer after substituting <code>&amp;, &lt;, &gt;,
     * </code>, and UTF-16 surrogates for the set of entitiy references
     * (<code>&amp;amp;, &amp;lt;, &amp;gt;</code>) and numeric character 
     * references (<code>&amp;#...</code>) respectively.
     * @param   writer      Writer for output.  Buffered writer is recommended.
     * @param   string      String to convert to XML format.
     * @param   encoding    CURRENTLY NOT IMPLEMENTED.
     */
    public void escapedWrite(Writer writer, String string, String encoding) throws IOException {
        //sun.io.CharToByteConverter ctbc = getCharToByteConverter(enc);
        for (int i = 0;  i < string.length();  i ++) {
            char ch = string.charAt(i);
            if ('<' == ch) {
                writer.write("&lt;");
            } else if ('>' == ch) {
                writer.write("&gt;");
            } else if ('&' == ch) {
                writer.write("&amp;");
                /**
            } else if ('"' == ch) {
                writer.write("&quot;");
            } else if ('\'' == ch) {
                writer.write("&apos;");
                */
            } else if (0xd800 <= ch && ch < 0xdc00) { // UTF-16 surrogate
                int next;
                if (i+1 >= string.length()) {
                    throw new RuntimeException("Invalid UTF-16 surrogate detected: "
                                               +Integer.toHexString(ch)+ " ?");
                } else {
                    next = string.charAt(++i);
                    if (!(0xdc00 <= next && next < 0xe000))
                        throw new RuntimeException("Invalid UTF-16 surrogate detected: "
                                                   +Integer.toHexString(ch)+" "
                                                   +Integer.toHexString(next));
                    next = ((ch-0xd800)<<10)+next-0xdc00+0x00010000;
                }
                writer.write("&#x");
                writer.write(Integer.toHexString(next));
                writer.write(";");
            /*} else if (null != ctbc && !ctbc.canConvert(ch)) {
                writer.write("&#x");
                writer.write(Integer.toString((int)ch, 16));
                writer.write(";");*/
            } else {
                writer.write(ch);
            }
        }
    }

    /**
     * Returns the specified <var>string</var> after substituting <VAR>specials</VAR>,
     * and UTF-16 surrogates for chracter references <CODE>&amp;#xnn</CODE>.
     *
     * @param   writer      Writer for output.  Buffered writer is recommended.
     * @param   string      String to convert to XML format.
     * @param   specials    Chracters, should be represeted in chracter referenfces.
     * @param   encoding    CURRENTLY NOT IMPLEMENTED.
     */
    public void escapedWrite(Writer writer, String string, String specials, String encoding)
        throws IOException {
        //sun.io.CharToByteConverter ctbc = getCharToByteConverter(enc);
        for (int i = 0;  i < string.length();  i ++) {
            char ch = string.charAt(i);
            int index = specials.indexOf(ch);
            if (index >= 0) {
                writer.write("&#");
                writer.write(Integer.toString(ch));
                writer.write(";");
            } else if (0xd800 <= ch && ch < 0xdc00) { // UTF-16 surrogate
                int next;
                if (i+1 >= string.length()) {
                    throw new RuntimeException("Invalid UTF-16 surrogate detected: "
                                               +Integer.toHexString(ch)+ " ?");
                } else {
                    next = string.charAt(++i);
                    if (!(0xdc00 <= next && next < 0xe000))
                        throw new RuntimeException("Invalid UTF-16 surrogate detected: "
                                                   +Integer.toHexString(ch)+" "
                                                   +Integer.toHexString(next));
                    next = ((ch-0xd800)<<10)+next-0xdc00+0x00010000;
                }
                writer.write("&#x");
                writer.write(Integer.toHexString(next));
                writer.write(";");
            /*} else if (null != ctbc && !ctbc.canConvert(ch)) {
                writer.write("&#x");
                writer.write(Integer.toString((int)ch, 16));
                writer.write(";");*/
            } else {
                writer.write(ch);
            }
        }
    }

    /**
     * Usage of sun.io package is not allowed in Pure Java.
     *
     * @param enc Java encoding name
     * @return CharToByteConverter an instance for enc, or null.
    protected sun.io.CharToByteConverter getCharToByteConverter(String enc) {
        sun.io.CharToByteConverter ctbc = null;
        if (null != enc) {
            try {
                ctbc = sun.io.CharToByteConverter.getConverter(enc);
            } catch (UnsupportedEncodingException e) {
            }
        }
        return ctbc;
    }
     */
}

