/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.sax;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Stack;
import java.util.Vector;
import org.xml.sax.AttributeList;
import org.xml.sax.DTDHandler;
import org.xml.sax.DocumentHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.helpers.LocatorImpl;

/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class NamespaceAdapter implements DocumentHandler, ErrorHandler {
    protected NamespaceDocumentHandler handler;
    protected ErrorHandler errorHandler = null;
    protected Stack elementStack;
    protected NamespaceAttributeListImpl attlist;
    protected Locator lastLocator = new LocatorImpl();
    protected Parser parser;

    /**
     * A helper method.
    public static Parser createParser(String classname,
                                      NamespaceDocumentHandler handler,
                                      ErrorHandler eHandler)
        throws ClassNotFoundException, IllegalAccessException, 
        InstantiationException, ClassCastException {

        Parser p = ParserFactory.makeParser(classname);
        NamespaceAdapter na = new NamespaceAdapter(handler);
        p.setDocumentHandler(na);
        p.setErrorHandler(na);
        na.setErrorHandler(eHandler);
        return p;
    }
     */

    /**
     *
     * @see org.xml.sax.helpers.ParserFactory#makeParser()
     */
    public NamespaceAdapter()
        throws java.lang.ClassNotFoundException,
        java.lang.IllegalAccessException,
        java.lang.InstantiationException,
        java.lang.ClassCastException {
        this(ParserFactory.makeParser());
    }

    /**
     *
     * @param saxParserName A class name of SAX Parser.
     * @see org.xml.sax.helpers.ParserFactory#makeParser(java.lang.String)
     */
    public NamespaceAdapter(String saxParserName)
        throws java.lang.ClassNotFoundException,
        java.lang.IllegalAccessException,
        java.lang.InstantiationException,
        java.lang.ClassCastException {
        this(ParserFactory.makeParser(saxParserName));
    }

    /**
     *
     */
    public NamespaceAdapter(Parser parser) {
        this.parser = parser;
        this.parser.setDocumentHandler(this);
        this.parser.setErrorHandler(this);
        this.elementStack = new Stack();
        this.attlist = new NamespaceAttributeListImpl(10);
    }

    // ----------------------------------------------------------------

    /**
     *
     */
    public void parse(InputSource source) throws SAXException, IOException {
        this.parser.parse(source);
    }

    /**
     *
     */
    public void parse(String systemId) throws SAXException, IOException {
        this.parser.parse(systemId);
    }

    // ----------------------------------------------------------------
    /**
     *
     */
    public void setDocumentHandler(NamespaceDocumentHandler handler) {
        this.handler = handler;
    }

    /**
     *
     */
    public void setDTDHandler(DTDHandler handler) {
        this.parser.setDTDHandler(handler);
    }

    /**
     *
     */
    public void setEntityResolver(EntityResolver resolver) {
        this.parser.setEntityResolver(resolver);
    }

    /**
     *
     */
    public void setErrorHandler(ErrorHandler handler) {
        this.errorHandler = handler;
    }

    /**
     *
     */
    public void setLocale(Locale locale) throws SAXException {
        this.parser.setLocale(locale);
    }

    // ----------------------------------------------------------------
    /**
     * Applications must not call this method.
     */
    public void startElement(String name, AttributeList atts)
        throws SAXException {
        ElementInfo einfo = new ElementInfo(name);
        for (int i = 0;  i < atts.getLength();  i ++) {
            String an = atts.getName(i);
            if (an.equals("xmlns")) {
                einfo.put("", atts.getValue(i));
            } else if (an.startsWith("xmlns:")) {
                einfo.put(an.substring(6).intern(), atts.getValue(i));
            }
        }
        this.elementStack.push(einfo);

        // Check namespace for element name
        int index = name.indexOf(':');
        if (index < 0) {
            einfo.prefix = null;
            einfo.localPart = name;
            einfo.namespace = this.searchForNamespace(""); // Search for default namespace
        } else {
                                                // PERFORMANCE TODO: Use String cache.
            einfo.prefix = name.substring(0, index).intern();
            einfo.localPart = name.substring(index+1).intern();
            if (einfo.prefix.length() == 0)
                this.error(new SAXParseException("No prefix: `"+name+"'.",
                                                 this.lastLocator));
            if (einfo.localPart.length() == 0)
                this.error(new SAXParseException("No localPart: `"+name+"'.",
                                                 this.lastLocator));
            einfo.namespace = this.searchForNamespace(einfo.prefix);
        }

        int alen = atts.getLength();
        this.attlist.removeAll();
        this.attlist.ensureCapacity(alen);
        for (int i = 0;  i < alen;  i ++) {
            String aname = atts.getName(i);
            int ind = aname.indexOf(':');
            if (ind < 0) {                      // per-element partition
                this.attlist.addAttribute(null, null, aname, aname,
                                          atts.getType(i), atts.getValue(i));
            } else {                            // global partition
                                                // PERFORMANCE TODO: Use String cache.
                String prefix = aname.substring(0, ind).intern();
                String localPart = aname.substring(ind+1).intern();
                if (prefix.length() == 0)
                    this.error(new SAXParseException("No prefix: `"+aname+"'.",
                                                     this.lastLocator));
                if (localPart.length() == 0)
                    this.error(new SAXParseException("No localPart: `"+aname+"'.",
                                                     this.lastLocator));
                this.attlist.addAttribute(this.searchForNamespace(prefix), prefix,
                                          localPart, aname, atts.getType(i),
                                          atts.getValue(i));
            }
        }
                                                // PERFORMANCE TODO: On-demand namespace resolution
        if (this.handler != null)
            this.handler.startElement(einfo.namespace, einfo.prefix, einfo.localPart,
                                      einfo.qname, this.attlist);
    }

    /**
     * Applications must not call this method.
     */
    public void endElement(String name)
        throws SAXException {
        if (this.elementStack.empty())
            return;                             // Ignore this event
        ElementInfo einfo = (ElementInfo)this.elementStack.peek();
        if (!einfo.qname.equals(name))
            return;                             // Ignore this event
        this.elementStack.pop();
        if (this.handler != null)
            this.handler.endElement(einfo.namespace, einfo.prefix, einfo.localPart, einfo.qname);
    }

    /**
     * @return null: No declaration for specified prefix or empty namespace
     */
    public String searchForNamespace(String prefix) throws SAXException {
        if (prefix.equals("xmlns"))
            return null;
        if (prefix.equals("xml"))
            return "http://www.w3.org/XML/1998/namespace";
        synchronized (this.elementStack) {
            for (int i = this.elementStack.size()-1;
                 i >= 0;  i --) {
                ElementInfo einfo = (ElementInfo)this.elementStack.elementAt(i);
                String namespace = einfo.get(prefix);
                if (namespace != null)
                    return namespace.length() == 0 ? null : namespace;
            }
        }
        if (prefix.length() == 0)
            return null;
        this.error(new SAXParseException("No namespace declaration for prefix `"+prefix+"'.",
                                         this.lastLocator));
        return null;
    }

    private static class ElementInfo {
        Hashtable xmlns;
        String qname;
        String prefix = null;
        String localPart;
        String namespace = null;

        ElementInfo(String name) {
            this.qname = name;
            this.xmlns = new Hashtable();
        }

        void put(String prefix, String namespace) {
            this.xmlns.put(prefix, namespace);
        }

        String get(String prefix) {
            return (String)this.xmlns.get(prefix);
        }
    }

    static class NamespaceAttributeListImpl implements NamespaceAttributeList {
        Vector values;

        public NamespaceAttributeListImpl(int len) {
            this.values = new Vector(len);
        }

        public int getLength() {
            return this.values.size();
        }

        public String getName(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).qname;
        }

        public String getType(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).type;
        }

        public String getValue(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).value;
        }

        public String getType(String name) {
            synchronized (this.values) {
                for (int i = 0;  i < this.values.size();  i ++) {
                    NamespaceAttribute nattr = (NamespaceAttribute)this.values.elementAt(i);
                    if (name.equals(nattr.qname))
                        return nattr.type;
                }
            }
            return null;
        }

        public String getValue(String name) {
            synchronized (this.values) {
                for (int i = 0;  i < this.values.size();  i ++) {
                    NamespaceAttribute nattr = (NamespaceAttribute)this.values.elementAt(i);
                    if (name.equals(nattr.qname))
                        return nattr.value;
                }
            }
            return null;
        }

        public String getPrefix(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).prefix;
        }

        public String getLocalPart(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).localPart;
        }

        public String getNamespace(int i) {
            if (i >= this.values.size())  return null;
            return ((NamespaceAttribute)this.values.elementAt(i)).namespace;
        }

        public void addAttribute(String namespace, String prefix, String localPart,
                                 String qname, String type, String value) {
            this.values.addElement(new NamespaceAttribute(namespace, prefix, localPart,
                                                          qname, type, value));
        }

        public void removeAll() {
            this.values.removeAllElements();
        }
        public void ensureCapacity(int cap) {
            this.values.ensureCapacity(cap);
        }
    }

    private static class NamespaceAttribute {
        String qname;
        String prefix;
        String localPart;
        String namespace;
        String value;
        String type;
        NamespaceAttribute(String n, String p, String l, String q, String t, String v) {
            this.qname = q;
            this.prefix = p;
            this.localPart = l;
            this.namespace = n;
            this.value = v;
            this.type = t;
        }
    }


    /**
     * Applications must not call this method.
     */
    public void setDocumentLocator(Locator locator) {
        this.lastLocator = locator;
        if (this.handler != null)
            this.handler.setDocumentLocator(locator);
    }
    /**
     * Applications must not call this method.
     */
    public void startDocument() throws SAXException {
        if (this.handler != null)
            this.handler.startDocument();
    }
    /**
     * Applications must not call this method.
     */
    public void endDocument() throws SAXException {
        if (this.handler != null)
            this.handler.endDocument();
    }
    /**
     * Applications must not call this method.
     */
    public void characters(char ch[], int start, int length)
        throws SAXException {
        if (this.handler != null)
            this.handler.characters(ch, start, length);
    }
    /**
     * Applications must not call this method.
     */
    public void ignorableWhitespace(char ch[], int start, int length)
        throws SAXException {
        if (this.handler != null)
            this.handler.ignorableWhitespace(ch, start, length);
    }
    /**
     * Applications must not call this method.
     */
    public void processingInstruction(String target, String data)
        throws SAXException {
        if (this.handler != null)
            this.handler.processingInstruction(target, data);
    }

    // ----------------------------------------------------------------
    /**
     * Applications must not call this method.
     */
    public void warning(SAXParseException exception) throws SAXException {
        if (this.errorHandler != null)
            this.errorHandler.warning(exception);
    }

    /**
     * Applications must not call this method.
     */
    public void error(SAXParseException exception) throws SAXException {
        if (this.errorHandler != null)
            this.errorHandler.error(exception);
    }

    /**
     * Applications must not call this method.
     */
    public void fatalError(SAXParseException exception) throws SAXException {
        if (this.errorHandler != null)
            this.errorHandler.fatalError(exception);
    }
    // ----------------------------------------------------------------
}
